#lang scheme/gui

(provide (all-defined-out))

;; bruce
;; needs:
;; a scale-number-of-dots function (of number of dots)
;; an upper-bound-number-of-dots function (of width, height, dot size)
;; "repulsion fields" implementation that keeps neighboring dots a minimal distance apart
;; upper-bound should take into consideration any "repulsion fields" in place,
;; because that limits the number of locations where a dot can be placed
;;
;; implemented scale-number-of-dots with an empirical formula
;;

;; changes: (bruce)
;; removed line that enables anti-aliasing (but still appears to be in effect)
;; changed < to <= in rectangles-overlap? and circles-overlap?
;; this was done to call dots that touch each other "overlapping"
;; added 1 to sum-of-radiuses in circles-overlap to round up to pixels (approximately)
;; that's because it does not represent an integral number of pixels due to square root usage
;;

;; A texture structure represents the properties of a texture that 
;; Bruce would like to use in generating mipmaps.
(define-struct texture
  (bg-color ; -> (is-a? arg color%)
   fg-color ; -> (is-a? arg color%)
   width ; -> (and (number? arg) (> arg 0) (even? arg) (< arg height))
   height  ; -> (and (number? arg) (> arg 0) (even? arg) (> arg width))
   min-dot-size ; -> (and (number? arg) (> arg 0) (< arg max-dot-size))
   max-dot-size ; -> (and (number? arg) (> arg 0) (> arg min-dot-size))
   dot-count ; -> (and (number? arg) (> arg 0))
   dot-shape ; -> (and (symbol? arg) (memq (list 'circle 'square)))
   debug-color ; (or (is-a? arg color%) (false? arg))
   dot-debug-color ; (or (is-a? arg color%) (false? arg))
   ))

;; A dot structure represents the properties of a shape that would
;; be drawn as a dot on a texture.
(define-struct dot 
  (x ; -> (and (number? arg) (>= arg 0))
   y ; -> (and (number? arg) (>= arg 0))
   size ; -> (and (number? arg) (>= arg 1))
   shape ; -> (and (symbol? arg) (memq (list 'circle 'square)))
   ))

;; A position represents the properties of a coordinate on an x y grid.
(define-struct posn
  (x ; -> (number? arg)
   y ; -> (number? arg)
   ))

;; texture->bitmap% : (texture? arg) -> (is-a? result bitmap%)
;; To render an in-memory bitmap% according to the properties specified
;; by the input texture.
(define (texture->bitmap% tdata)
  (let ((bitmap (make-object bitmap% 
                  (texture-width tdata)
                  (texture-height tdata)))
        (dc (new bitmap-dc%)))
    (begin
      ;; connect the bitmap to the dc
      (send dc set-bitmap bitmap)   
      ;; enable anti-aliasing

      
      ;; **********************************************************************
      ;; **********************************************************************
              ;;(send dc set-smoothing 'aligned)
      ;; **********************************************************************
      ;; **********************************************************************

      
      
      ;; configure the drawing pen and brush for the background
      (send dc set-pen 
            (make-object pen% (texture-bg-color tdata) 1 'solid))
      (send dc set-brush 
            (make-object brush% (texture-bg-color tdata) 'solid))
      ;; draw the solid background color
      (send dc draw-rectangle 0 0 (texture-width tdata) (texture-height tdata))
      ;; configure the drawing pen and brush for the foreground
      ;; for the dot shapes
      (let* ((dbg-pen (texture-dot-debug-color tdata))
             (normal-pen (texture-fg-color tdata))
             (the-pen 
              (make-object pen% 
                (if dbg-pen dbg-pen normal-pen) 1 'solid)))
        (send dc set-pen the-pen))
      (send dc set-brush 
            (make-object brush% (texture-fg-color tdata) 'solid))
      ;; start drawing the dot shapes
      (draw-shapes dc tdata)
      ;; draw the texture debug if necessary
      (debug-texture tdata dc)
      ;; disconnect the bitmap from the dc
      (send dc set-bitmap #f)
      bitmap)))

;; debug-texture : (texture? arg) (implementation? arg dc<%>)
;; -> void
;; To draw an outline over the texture to help verify that is being
;; drawn in the correct location.
(define (debug-texture tdata dc)
  (let ((debug-color (texture-debug-color tdata)))
    (when debug-color
      (let ((dbg-pen (make-object pen% debug-color 1 'solid))
            (dbg-brush (make-object brush% 
                         debug-color 'transparent)))
        (send dc set-pen dbg-pen)
        (send dc set-brush dbg-brush)
        (send dc draw-rectangle 0 0 
              (texture-width tdata) (texture-height tdata))))))

;; draw-shapes : (implementation arg dc<%>) (texture? arg)
;; -> (void? result)
;; To randomly draw shapes on a dc<%>.
;; Shapes are drawn in random locations,
;; but do not overlap or exceed the bounds of the texture. The
;; size of the shapes are random, bound by the specified minimum and
;; maximum shape size.
;; [---------------------------- shape size ---------------------]
;; (texture-min-dot-size tdata)
;;                                    (texture-min-dot-size tdata)
;;
;; Shapes are drawn given the coordinates for their top-left corner
;; [------------ x -----------------------------------------------]
;; 0                        (- (texture-width tdata) [shape size]))
;;
;; [------------ y -----------------------------------------------]
;; 0                       (- (texture-height tdata) [shape size]
(define (draw-shapes dc tdata)
  (let loop ((dot-count (texture-dot-count tdata)) (drawn-dots '()))
    (cond
      ((zero? dot-count) (void))
      (else
       (let* ((x (random (texture-width tdata)))
              (y (random (texture-height tdata)))
              (width (texture-width tdata))
              (height (texture-height tdata))
              (shape-size (random-size tdata))
              (dot-shape (texture-dot-shape tdata))
              (the-dot (make-dot x y shape-size dot-shape)))
         (if (and (in-bounds? x y width height shape-size)
                  (not (overlapped? the-dot drawn-dots)))
             (begin
               (draw-dot the-dot dc)
               (loop (sub1 dot-count) 
                     (cons the-dot
                           drawn-dots)))
             (loop dot-count drawn-dots)))))))

;; draw-dot : (dot? arg) -> void
;; To draw a dot according to its properties
(define (draw-dot the-dot dc)
  (let ((x (dot-x the-dot))
        (y (dot-y the-dot))
        (shape-size (dot-size the-dot))
        (dot-shape (dot-shape the-dot)))
    (case dot-shape
      ((circle) 
       (send dc draw-ellipse x y shape-size shape-size))
      ((square)
       (send dc draw-rectangle x y shape-size shape-size))
      (else 
       (error "Unknown shape type" dot-shape)))))

;; overlapped? : (dot? arg) 
;; (listof (dot? arg)) -> (boolean? result)
;; To determine whether the two dot shapes, if rendered on a
;; texture, would overlap.
(define (overlapped? this-dot drawn-dots)
  (case (dot-shape this-dot)
    ((circle) (ormap (lambda (that-dot) 
                       (circles-overlap? this-dot that-dot))
                     drawn-dots))
    ((square) (ormap (lambda (that-dot) 
                       (rectangles-overlap? this-dot that-dot))
                     drawn-dots))
    (else (error "Unknown shape type"))))

;; rectangles-overlap? : 
;; (and (dot? arg) (symbol=? (dot-shape arg) 'square))
;; (and (dot? arg) (symbol=? (dot-shape arg) 'square))
;;  -> (boolean? result)
;; To determine whether two rectangles would overlap
;; if they were drawn on the same texture. 
;; Translated from:
;; http://compsci.ca/v3/viewtopic.php?t=13661
;; For a single rectangle, use the following properties:
;; * x1: lowest x value of the rectangle 
;; * y1: lowest y value of the rectangle 
;; * x2: greatest x value of the rectangle 
;; * y2: greatest y value of the rectangle
;; Two rectangles overlap if:
;; * box #1's x1 (lowest value) has to be LESS than box #2's x2 
;;   (greatest value) 
;; * box #1's x2 (greatest value) has to be GREATER than box #2's x1
;;   (lowest value) 
;; * box #1's y1 (lowest value) has to be LESS than box #2's y2 
;;   (greatest value) 
;; * box #1's y2 (greatest value) has to be GREATER than box #2's y1 
;;   (lowest value)
(define (rectangles-overlap? square1 square2)
  (let* ((sq1-size (dot-size square1))
         (sq1-x1 (dot-x square1))
         (sq1-y1 (dot-y square1))
         (sq1-x2 (+ sq1-x1 sq1-size))
         (sq1-y2 (+ sq1-y1 sq1-size))
         (sq2-size (dot-size square2))
         (sq2-x1 (dot-x square2))
         (sq2-y1 (dot-y square2))
         (sq2-x2 (+ sq2-x1 sq2-size))
         (sq2-y2 (+ sq2-y1 sq2-size)))
    (and (<= sq1-x1 sq2-x2)
         (>= sq1-x2 sq2-x1)
         (<= sq1-y1 sq2-y2)
         (>= sq1-y2 sq2-y1))))

;;
;;    (and (<= sq1-x1 (+ sq2-x2 1))
;;         (>= sq1-x2 (- sq2-x1 1))
;;         (<= sq1-y1 (+ sq2-y2 1))
;;         (>= sq1-y2 (- sq2-y1 1)))))


;; bruce
;; changed < to <= etc to count edges touching as overlapping
;; add + 1 and - 1 to create a buffer or "repulsion field"
;; (carefully)
;;
;; using 1 causes problems so remove that and use a fraction of the dot size
;; problem was it couldn't find places for enough dots on smaller images, had to scale down #dots
;; that made larger images too sparse to look good
;; need to add a feature that gives up looking for a place for more dots after a few attempts
;; then this could change again
;;

;; in-bounds? : (number? arg) (number? arg) (number? arg) 
;; (number? arg) (number? arg) -> (boolean? result)
;; To determine whether the x and y coordinate for the shape
;; are within the bounds of the texture.
(define (in-bounds? x y width height shape-size)
  (and (valid-x x width shape-size) 
       (valid-y y height shape-size)))

;; valid-x : (number? arg) (number? arg) (number? arg) 
;; -> (boolean? result)
;; To determine whether the x coordinate would allow the shape
;; to be drawn entirely on the canvas.
;; (valid-x 10 20 10) -> #t
;; (valid-x 10 19 10) -> #f
(define (valid-x x width shape-size)
  (<= x (- width shape-size)))

;; valid-y : (number? arg) (number? arg) (number? arg) 
;; -> (boolean? result)
;; To determine whether the y coordinate would allow the shape
;; to be drawn entirely on the canvas.
;; (valid-y 10 20 10) -> #t
;; (valid-y 10 19 10) -> #f
(define (valid-y y height shape-size)
  (<= y (- height shape-size)))

;; random-size : (texture? arg) -> (number? result)
;; To randomly generate a number within a given range:
;;                            [---------- result -----------]
;; (texture-min-dot-size tdata) 
;;                               (texture-max-dot-size tdata)
(define (random-size tdata)
  (let* ((min (texture-min-dot-size tdata))
         (max (texture-max-dot-size tdata))
         (the-range (range min max)))
    (+ (random the-range) min)))

;; range : (and (number? arg) (< arg the-max)) 
;;         (and (number? arg) (> arg the-min)) -> (number? result)
;; To determine the number of numbers included in the interval
;; specified.
;;       [12345................result]
;; the-min                     the-max
;; (range 2 4) -> 3
;; (range 3 7) -> 5
(define (range the-min the-max)
  (add1 (- the-max the-min)))

;; string->color% : (string? arg) -> (is-a? result color%)
;; To find a color% by name.
;; Delegates work to the the-color-database, of type color-database<%>.
;; (string->color% "Tomato") -> color%
;; (string->color% "Violet") -> color%
;; (string->color% "Sienna") -> color%
;; (string->color% "Chocolate") -> color%
;; See the color-database<%> documentation for a list of color names.
(define (string->color% name)
  (send the-color-database find-color name))

;; bitmap%->image-snip% : 
;; (is-a? arg bitmap%) -> (is-a? result image-snip%)
;; To create a snip configurd to display the contents of the bitmap%.
;; image-snip%s can be displayed in the definitions windows (REPL).
(define (bitmap%->image-snip% bitmap)
  (let ((snip (make-object image-snip%)))
    (begin
      (send snip set-bitmap bitmap)
      snip)))

;; image-snip%-to-png : 
;; (is-a? arg image-snip%) (string? arg) -> (boolean? result)
;; Save an image-snip% to a PNG file.
(define (image-snip%-to-png snip file-name)
  (send (send snip get-bitmap) save-file file-name 'png))



;; bruce
;; my attempt at scale-number-of-dots
;; the smaller the dots get, the more of them I want
;; can use width and height to calculate a maximum possible number of dots
;; width * height / max-dot-size or similar would be the maximum
(define (scale-number-of-dots last-number-of-dots max-dot-size width height)
;;  (ceiling (* 1/2 last-number-of-dots)))
;;  (ceiling (* 1/4 last-number-of-dots)))

  ;; assume it is safe to divide by max-dot-size, constrained elsewhere
  ;; actually divide by max-dot-size squared * 2 squared and an extra 2 for good measure
  (let*
    ((max-number-of-dots-for-size  (floor (/ (* width height) (* max-dot-size max-dot-size 8))))
     (new-approximation   (ceiling (* 1/2 last-number-of-dots)))
     (result   (min new-approximation max-number-of-dots-for-size)))
     (printf "~n*****~nscale-number-of-dots:~n")
     (printf "last-number-of-dots ....... = ~a~n" last-number-of-dots)
     (printf "max-dot-size .............. = ~a~n" max-dot-size)
     (printf "width ..................... = ~a~n" width)
     (printf "height .................... = ~a~n~n" height)
     (printf "new-approximation ......... = ~a~n" new-approximation)
     (printf "max-number-of-dots-for-size = ~a~n" max-number-of-dots-for-size)
     (printf "result .................... = ~a~n*****~n" result)
     result))


;; next-texture : (texture? arg) -> (texture? result)
;; To create the next texture by reducing width and height by
;; 1/2, the dot count by 1/2, and the max dot size by 1/2.
;; Since width and height are required to be even, they can keep
;; getting reduced as the fractions are always multiples of two.
;; max-dot-size and dot-count, however, are not required to be even.
;; As such, their result is always rounded up to the next whole value
;; using the ceiling function. Consequently, they will never have a
;; value less than 1; which is important to assure that something will
;; be drawn on a texture.
(define (next-texture tdata)
  (let* 
    ((new-width          (* 1/2 (texture-width tdata)))
     (new-height         (* 1/2 (texture-height tdata)))
     (new-max-dot-size   (ceiling (* 1/2 (texture-max-dot-size tdata)))))
  (make-texture
   (texture-bg-color tdata)
   (texture-fg-color tdata)
   new-width
   new-height
   (texture-min-dot-size tdata)
   new-max-dot-size
   (scale-number-of-dots (texture-dot-count tdata)
                         new-max-dot-size
                         new-width
                         new-height)
   (texture-dot-shape tdata)
   (texture-debug-color tdata)
   (texture-dot-debug-color tdata))))

;; next-textures : (texture? arg) (number? arg) -> (listof texture?)
;; To create a list composed of the first texture, and the next n
;; textures that follow.
(define (next-textures tdata n)
  (cons tdata
        (let loop ((tdata tdata) (n n))
          (cond
            ((zero? n) empty)
            (else
             (let ((new-texture (next-texture tdata)))
               (cons new-texture 
                     (loop new-texture (sub1 n)))))))))

;; display-texture ; (texture? arg) -> void
;; To display texture information on the Interactions windows.
(define (display-texture tdata)
  (printf "~a~n~a~n~a~n~a~n~a~n~a~n~a~n~a~n~a~n~a~n~n"
          (texture-bg-color tdata)
          (texture-fg-color tdata)
          (texture-width tdata)
          (texture-height tdata)
          (texture-min-dot-size tdata)
          (texture-max-dot-size tdata)
          (texture-dot-count tdata)
          (texture-dot-shape tdata)
          (texture-debug-color tdata)
          (texture-dot-debug-color tdata)))

;; textures->bitmap% : (listof (texture? arg))
;; -> (is-a? result bitmap%)
;; To render an in-memory bitmap% composed of the provided textures
;; to create a "mipmap".
(define (textures->bitmap% textures)
  (let* ((tdata (car textures))
         (width (apply + (map texture-width textures)))
         (height (texture-height tdata))
         (bitmaps (map (λ (texture) (texture->bitmap% texture))
                       textures))
         (bitmap (make-object bitmap% width height))
         (dc (new bitmap-dc%))
         (bg-pen (make-object pen% (texture-bg-color tdata) 1 'solid))
         (bg-brush 
          (make-object brush% (texture-bg-color tdata) 'solid)))
    (begin
      (send dc set-bitmap bitmap)
      ;; draw the solid background color
      (send dc set-pen bg-pen)
      (send dc set-brush bg-brush)
      (send dc draw-rectangle 0 0 width height)
      ;; copy in the textures bitmaps
      (let loop ((x 0) (y 0) (bitmaps bitmaps))
        (if (null? bitmaps) (void)
            (let ((bitmap (car bitmaps)))
              (begin
                (send dc draw-bitmap bitmap x y)
                (loop (+ x (send bitmap get-width)) 
                      y 
                      (cdr bitmaps))))))
      (send dc set-bitmap #f)
      bitmap)))

;; texture->mipmap : (texture? arg) (string? arg) -> (boolean? result)
;; To create a PNG mipmap per Bruce's spec.
(define (texture->mipmap texture file-name)
  (let* ((textures (next-textures texture 4))
         (bitmap (textures->bitmap% textures))
         (file-name (string-append 
                     file-name                     
                     (number->string
                      (current-seconds))
                     ".png")))
    (send bitmap save-file file-name 'png)))

;; display-dot : (dot? arg) -> void
;; To display dot information on the Interactions window.
(define (display-dot the-dot)
  (printf "dot~nx: ~a~ny: ~a~nsize: ~a~nshape: ~a~n" 
          (dot-x the-dot)
          (dot-y the-dot)
          (dot-size the-dot)
          (dot-shape the-dot)))

;; distance : (posn? arg) (posn? arg) -> (number? result)
;; To determine the distance between two points on a coordinate grid
;; using the Pythagorean theorem: a^2 + b^2 = c^2
(define (distance posn1 posn2)
  (let ((a (- (posn-x posn2) (posn-x posn1)))
        (b (- (posn-y posn2) (posn-y posn1))))
    (sqrt (+ (expt a 2)
             (expt b 2)))))

;; circles-overlap? :
;; (and (dot? arg) (symbol=? (dot-shape arg) 'circle))
;; (and (dot? arg) (symbol=? (dot-shape arg) 'circle))
;;  -> (boolean? result)
;; To determine whether two circles overlap. ThiS DOES NOT WORK for ellipses
;; other than circles.
;; Translated from:
;; http://compsci.ca/v3/viewtopic.php?t=13661
;; We need:
;; * co-ordinates of the centre point (radii) (x and y) of both circles 
;; * radius of both circles
;;
;; There are three steps to check for a collision between two circles:
;; 1. Find the distance between the centres of the two circles. 
;; 2. Add the two radii of the circles 
;;    (radius of one circle + radius of the other circle). 
;; 3. Check if the distance between the centres of the circles is less 
;;    than the sum of the two radii and, if it is, then the two circles 
;;    have collided or are overlapping.
(define (circles-overlap? circle1 circle2)
  (let* ((circle1-radii (radii-of-circle circle1))
         (circle2-radii (radii-of-circle circle2))
         (distance-between-circles (distance circle1-radii circle2-radii))
         (circle1-radius (* 1/2 (dot-size circle1)))
         (circle2-radius (* 1/2 (dot-size circle2)))         
         (sum-of-radiuses (+ circle1-radius circle2-radius)))
    (<= distance-between-circles (+ 1 sum-of-radiuses))))

;; bruce
;; changed < to <= and added 1 to sum of radiuses to create a larger area considered as overlapping
;; i.e. trying to keep some space between edges, first they overlap if they touch, then add extra
;; did the same/similar with rectangeles-overlap?
;;
;; could use a little work on the terminology:
;; radius is singular, radii is plural
;; radii-of-circle below really returns the center point of a circular dot
;; should be called center-point-of-circle
;;
;; I think circles-overlap? should use ceil to round up the radii
;; to integral values that represent pixels,
;; otherwise two edges of the same pixel would not be considered overlapping,
;; but they clearly are overlapping, they are the same pixel!
;;
;; that causes difficulty if the radii are less than 1 in size but it should still work!
;; adding 1 to the sum of radii is a hack that approximates this correction
;;
;; values are not integral here,
;; because the square root value can clearly be non-integral
;;

;; radii-of-circle : (symbol=? (dot-shape arg) 'circle) 
;; -> (posn? result)
;; To determine the radii of the circle.
(define (radii-of-circle circle)
  (make-posn (+ (dot-x circle)
                (/ (dot-size circle) 2))
             (+ (dot-y circle)
                (/ (dot-size circle) 2))))